package GuitarKey.Filters;

import java.util.HashMap;
import java.util.Iterator;
import java.util.logging.Logger;

import GuitarKey.Utils.ApplicationModes;
import GuitarKey.Utils.Scale;
import GuitarKey.Utils.ScaleLocation;

/**
 * Class filters scales for displaying purposes
 */
public class ScaleFilter 
{
	// Logger
	private static Logger logger = Logger.getLogger("GuitarKey");
	
	// Different Filtering Schemes
	public static final int NO_FILTER = 0;
	public static final int THREE_NOTE_FILTER = 1;
	public static final int BOX_FILTER = 2;
	
	private int filterMode = NO_FILTER;
	
	// Position determines which note is the root
	int position = 0;
	
	Scale scale = null;
	
	private HashMap positionMap;
	
	/**
	 * Default Constructor
	 */
	public ScaleFilter()
	{
		buildPositionMap();
	}
	
	private void buildPositionMap()
	{
		// Build the positions map
		positionMap = new HashMap(7);
		
		positionMap.put("I", 0);
		positionMap.put("II", 1);
		positionMap.put("III", 2);
		positionMap.put("IV", 3);
		positionMap.put("V", 4);
		positionMap.put("VI", 5);
		positionMap.put("VII", 6);
	}
	
	/**
	 * Set the filtering position
	 */
	public void setFilterPosition(String inPosition)
	{
		if (positionMap.containsKey(inPosition))
		{
			// Grab the position offset from the map
			position = ((Integer)positionMap.get(inPosition)).intValue();
		}
	}
	
	/**
	 * Apply the filter to the scale
	 */
	public void filter(Scale inScale) 
	{
		scale = inScale;
		
		filterMode = ApplicationModes.filterMode;
		
		// Switch filtering schemes
		switch(filterMode)
		{
			case NO_FILTER:
				applyDefaultFilter();
				break;
			case THREE_NOTE_FILTER:
				apply3NoteFilter();
				break;
			case BOX_FILTER:
				applyBoxFilter();
				break;
			default:
				applyDefaultFilter();
				break;
		}
	}
	
	/**
	 * Apply the default (no)filter to the scale and set the alpha values
	 */
	private void applyDefaultFilter()
	{
		Iterator itrScaleLocations = scale.getScaleLocations().iterator();
		while (itrScaleLocations.hasNext())
		{
			ScaleLocation tmpScaleLoc = (ScaleLocation)itrScaleLocations.next();
			
			tmpScaleLoc.setAlpha(false);	
		}
	}
	
	/**
	 * Apply the 3 Note filter to the scale and set the alpha values
	 */
	private void apply3NoteFilter()
	{
		logger.info("Applying filter 3 Note Per String Filter in position " + position);
		
		int lastStringNum = 1;
		int stringNum = 0;
		int idx = 0;
		int index = 0;
		
		String strNote ="";
		String strLastNote = "";
		
		scale.sortScaleLocations();
		
		// Get the scale note names array
		// and find the root note index
		String[] noteNames = scale.getNoteNameArray();
		strNote = scale.getScaleRoot();
		for (int i=0; i<noteNames.length; i++)
		{
			if (noteNames[i].equalsIgnoreCase(strNote))
			{
				idx = i + position;
				
				if (idx >= noteNames.length)
				{
					idx -= noteNames.length;
				}
			}
		}
		
		Iterator itrScaleLocations = scale.getScaleLocations().iterator();
		while (itrScaleLocations.hasNext())
		{
			ScaleLocation tmpScaleLoc = (ScaleLocation)itrScaleLocations.next();
			
			tmpScaleLoc.setAlpha(true);
			strNote = tmpScaleLoc.getNoteName();
			stringNum = tmpScaleLoc.getStringLocation();
			
			if (lastStringNum != stringNum)
			{
				// Increment Note Names index by 3 when switching to
				// a new string
				idx += 3;
				if (idx >= noteNames.length)
				{
					idx -= noteNames.length;
				}
			}
			
			index = idx;
			for (int j=0; j<3; j++)
			{
				if (index >= noteNames.length)
				{
					index = 0;
				}
				
				// Check to see if scale location is equal to one of the 3 notes
				// for this string
				if (strNote.equalsIgnoreCase(noteNames[index]))
				{
					tmpScaleLoc.setAlpha(false);
				}
				
				index++;
			}
			
			lastStringNum = stringNum;
		}
	}
	
	/**
	 * Apply the Box Note filter to the scale and set the alpha values
	 */
	private void applyBoxFilter()
	{
		logger.info("Applying Box Pattern Filter in position " + position);
		
		int fretLoc = 0;
		int rootFret = 0;
		boolean foundRoot = false;
		int idx = 0;
		
		int lowBound = 0;
		int highBound = 0;
		
		String strNote = "";
		
		scale.sortScaleLocations();
		
		// Find the root note index in the note names array
		String[] noteNames = scale.getNoteNameArray();
		strNote = scale.getScaleRoot();
		for (int i=0; i<noteNames.length; i++)
		{
			if (noteNames[i].equalsIgnoreCase(strNote))
			{
				idx = i + position;
				
				if (idx >= noteNames.length)
				{
					idx -= noteNames.length;
				}
			}
		}
		
		Iterator itrScaleLocations = scale.getScaleLocations().iterator();
		while (itrScaleLocations.hasNext())
		{
			ScaleLocation tmpScaleLoc = (ScaleLocation)itrScaleLocations.next();
			
			tmpScaleLoc.setAlpha(true);
			fretLoc = tmpScaleLoc.getFretLocation();
			strNote = tmpScaleLoc.getNoteName();
			
			if (!foundRoot)
			{
				// Find root note
				if (noteNames[idx].equalsIgnoreCase(strNote))
				{
					foundRoot = true;
					rootFret = fretLoc;
					tmpScaleLoc.setAlpha(false);
					
					// Set the higher and upper bounds
					lowBound = rootFret - 1;
					highBound = rootFret + 3;
					
					if (highBound >= ApplicationModes.NUMBER_OF_NOTES_IN_OCTAVE)
					{
						highBound -= ApplicationModes.NUMBER_OF_NOTES_IN_OCTAVE;
					}
					if (lowBound < 0)
					{
						lowBound += ApplicationModes.NUMBER_OF_NOTES_IN_OCTAVE;
					}
				}
			}
			else
			{
				// If Scale Location is within bounds, display it
				if ((fretLoc >= lowBound) && (fretLoc <= highBound))
				{
					tmpScaleLoc.setAlpha(false);
				}	
			}
		}
		
		// High bound is lower than low bound when box crosses 12th fret
		if (highBound < lowBound)
		{
			itrScaleLocations = scale.getScaleLocations().iterator();
			while (itrScaleLocations.hasNext())
			{
				ScaleLocation tmpScaleLoc = (ScaleLocation)itrScaleLocations.next();
				
				fretLoc = tmpScaleLoc.getFretLocation();
				strNote = tmpScaleLoc.getNoteName();
				
				if ((fretLoc >= lowBound) || (fretLoc <= highBound))
				{
					tmpScaleLoc.setAlpha(false);	
				}
			}
		}
	}
}
