package GuitarKey.Gui;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JPanel;

import GuitarKey.Filters.ScaleFilter;
import GuitarKey.Instruments.Fretboard;
import GuitarKey.Instruments.Guitar;
import GuitarKey.Instruments.FretboardLocationMap.Orientation;
import GuitarKey.Sound.MidiPlayer;
import GuitarKey.Utils.ApplicationModes;
import GuitarKey.Utils.HelpLauncher;
import GuitarKey.Utils.Scale;
import GuitarKey.Utils.ScaleManager;
import GuitarKey.Utils.TuningsManager;

/**
 * Main Guitar Key App.
 *
 * @author Devon Bryant
 * @version 1.2.0 05/21/2006
 */
public class GuitarKeyFrame extends JFrame implements ActionListener 
{
	// Logger
	private static Logger logger = Logger.getLogger("GuitarKey");
	
	private ApplicationModes appModes;
	private ButtonMap buttonMap;
	private ScaleManager scaleManager;
	private TuningsManager tuningsManager;
	private MidiPlayer midiPlayer;
	private HelpLauncher helpLauncher;
	
	ArrayList<Guitar> guitarList;
	
	private GuitarPanel panel;
	private GuitarMenuBar guitarMenu;
	
	private Scale blankScale;
	
	/**
     * Constructor.
     */  
     public GuitarKeyFrame() 
     {
     	logger.info("Creating GuitarKeyFrame");
     	
     	Toolkit tk = Toolkit.getDefaultToolkit();
     	Dimension screenSize = tk.getScreenSize();
        int screenWidth = (screenSize.width - 798) / 2;
     	
     	// Setup the main components
     	appModes = new ApplicationModes();
     	scaleManager = new ScaleManager();
     	tuningsManager = new TuningsManager();
     	midiPlayer = new MidiPlayer();
     	helpLauncher = new HelpLauncher();
     	guitarMenu = new GuitarMenuBar(this);
        
        // Load Scale Menu
        java.util.List scaleList = scaleManager.getScaleNames();
    	Iterator itrScaleNames = scaleList.iterator();
    	
    	// Cycle through the scale names
    	while (itrScaleNames.hasNext())
    	{
    		guitarMenu.addScaleName((String)itrScaleNames.next());
    	}
    	
    	// Load Tuning Menu
        java.util.List tuningsList = tuningsManager.getTuningNames();
    	Iterator itrTuningsNames = tuningsList.iterator();
    	
    	// Cycle through the tuning names
    	while (itrTuningsNames.hasNext())
    	{
    		guitarMenu.addTuningName((String)itrTuningsNames.next());
    	}
        
        
        setTitle("Guitar Scale Assistant");
        setIconImage(new ImageIcon(getClass().getClassLoader().getResource("img/icon.gif")).getImage());
        setBackground(new Color(212, 208, 200));
        setJMenuBar(guitarMenu);
        setResizable(false);
        
        // Draw panel
        panel = new GuitarPanel();
        panel.setBounds(0, 0, 798, 623);
        panel.setLayout(null);
        panel.setBackground(new Color(212, 208, 200));
        
     	
     	// Button Map
     	buttonMap = new ButtonMap(this);
     	panel.add(buttonMap);
     	
     	// Guitar List
      Orientation orientation = Orientation.LEFT_HAND;
     	buildGuitarList(orientation);
     	
     	// Blank scale
		blankScale = new Scale("C", "Unknown");
     	
        // Add window listener.
        this.addWindowListener
        (
            new WindowAdapter() {
                public void windowClosing(WindowEvent e) {
                    GuitarKeyFrame.this.windowClosed();
                }
            }
        ); 
        
        setContentPane(panel);
        pack();
     	setSize(new Dimension(798, 623));
     	setLocation(screenWidth, 0);
     	
     	setVisible(true); // Show App
     	
    }

    
    /**
     * Method builds the guitar list
     */
    private void buildGuitarList(Orientation orientation)
    {
    	guitarList = new ArrayList<Guitar>();
    	
    	// Top Guitar
    	Guitar topGuitar = new Guitar(Fretboard.CLICKABLE, orientation);
    	topGuitar.setLocation(appModes.LEFT_SHIFT, appModes.TOP_GUIT_LOCATION);
    	topGuitar.setVisible(true);
    	guitarList.add(topGuitar);
    	
    	// Bottom Guitar
    	Guitar bottomGuitar = new Guitar(Fretboard.DISPLAY, orientation);
    	bottomGuitar.setLocation(appModes.LEFT_SHIFT, appModes.BOTTOM_GUIT_LOCATION);
    	bottomGuitar.setVisible(true);
    	guitarList.add(bottomGuitar);
    	
      // Add Guitars to panel
    	panel.add(topGuitar);
    	panel.add(bottomGuitar);
    }
    
    /**
     * Method handles all actions from menu and buttons
     */
    @SuppressWarnings("deprecation")
    private void handleAction(String inAction)
    {
    	buttonMap.hideInfo();
    	
    	// Clear the Fretboard
    	if (inAction.equals("Clear Scale"))
        {
        	logger.info("Clearing the clickable fretboard");
        	Iterator itrGuitars = guitarList.iterator();
        	while (itrGuitars.hasNext())
        	{
        		((Guitar)itrGuitars.next()).clearNotes();
        	}
        }
        // Clear the bottom display Fretboard
    	if (inAction.equals("Clear Display"))
        {
        	logger.info("Clearing the display fretboard");
        	Iterator itrGuitars = guitarList.iterator();
        	while (itrGuitars.hasNext())
        	{
        		Fretboard tmpFretboard = ((Guitar)itrGuitars.next()).getFretboard();
        		if (tmpFretboard.getFretboardType() == Fretboard.DISPLAY)
        		{
        			tmpFretboard.setScale(blankScale);
        		}
        	}
        }
        // Show the save interface
        else if (inAction.equals("Save"))
        {
        	logger.info("Initialize scale save");
        	buttonMap.showSaveInterface();
        }
        // Show the save tuning interface
        else if (inAction.equals("Save Tuning"))
        {
        	logger.info("Initialize tuning save");
        	buttonMap.showSaveTuningInterface();
        }
        // Hide the save interface
        else if (inAction.equals("Cancel Save"))
        {
        	logger.info("Cancel save");
        	buttonMap.hideSaveInterface();
        	buttonMap.hideSaveTuningInterface();
        }
        // Save the scale
    	else if (inAction.equals("Save Scale"))
        {
        	logger.info("Attempting to save scale");
        	Iterator itrGuitars = guitarList.iterator();
        	while (itrGuitars.hasNext())
        	{
        		Guitar tmpGuitar = (Guitar)itrGuitars.next();
        		Fretboard tmpFretboard = tmpGuitar.getFretboard();
        		if (tmpFretboard.getFretboardType() == Fretboard.CLICKABLE)
        		{
        			if (tmpFretboard.hasScaleLocations())
        			{
        				// Get the current tuning of the guitar
        				ArrayList tuningShifts = tmpGuitar.getTuningShifts();
        				
        				// Reset Guitar to standard tuning first
        				tmpGuitar.resetToStandard(false);
        				
        				// Sort the scale before saving
        				Scale tmpScale = tmpFretboard.getScale();
        				tmpScale.sortScaleLocations();
        			
        				// Set the scale name and root note
        				tmpScale.setScaleName(buttonMap.getScaleName());
        				tmpScale.setScaleRoot(buttonMap.getScaleRoot());
        			
        				// Add Scale to the Scale Manager and file menu if it doesn't already exist
        				boolean containsScale = scaleManager.addScale(tmpScale.clone(), true);
        				if (!containsScale)
        				{
        					guitarMenu.addScaleName(tmpScale.getScaleFullName());
        					logger.info("Saved " + tmpScale.getScaleFullName() + " scale");
        				}
        				else
        				{
        					logger.log(Level.WARNING, "Scale Already Exists");
        				}
        				
        				// Retune the guitar to the previous tuning
        				tmpGuitar.tuneClickableGuitar(tuningShifts);
        				
        				// Show Info
        				buttonMap.showInfo("Scale Saved");
        			}
        			else
        			{
        				logger.log(Level.WARNING, "Attempted to save empty scale");
        				// Show No Notes selected info
        				buttonMap.showInfo("No Notes");
        			}
        		}
        	}
        	
        	// Hide Interface
        	buttonMap.hideSaveInterface();
        }
        // Save the tuning
    	else if (inAction.equals("Save New Tuning"))
        {
        	logger.info("Attempting to save tuning");
        	Iterator itrGuitars = guitarList.iterator();
        	while (itrGuitars.hasNext())
        	{
        		Guitar tmpGuitar = (Guitar)itrGuitars.next();
        		Fretboard tmpFretboard = tmpGuitar.getFretboard();
        		if (tmpFretboard.getFretboardType() == Fretboard.CLICKABLE)
        		{
    				// Get the current tuning of the guitar
    				ArrayList tuningShifts = tmpGuitar.getTuningShifts();
    				
    				String tuningName = buttonMap.getTuningName();
    				
    				if ((tuningName == null) || (tuningName.length() == 0))
    				{
    					// No tuning name was specified
    					buttonMap.showInfo("No Name");
    				}
    				else
    				{
	    				// Add Tuning to the Tuning Manager and file menu if it doesn't already exist
	    				boolean containsTuning = tuningsManager.addTuning(tuningName, tuningShifts, true);
	    				if (!containsTuning)
	    				{
	    					guitarMenu.addTuningName(tuningName);
	    					logger.info("Saved " + tuningName + " tuning");
	    				}
	    				else
	    				{
	    					logger.log(Level.WARNING, "Tuning Already Exists");
	    				}
	    				
	    				// Show Info
	    				buttonMap.showInfo("Tuning Saved");
	    			}
        		}
        	}
        	
        	// Hide Interface
        	buttonMap.hideSaveTuningInterface();
        }
        // Tune Guitars to standard tuning
        else if (inAction.equals("Tune Standard"))
        {
        	logger.info("Reseting guitars to standard tuning");
        	
        	Iterator itrGuitars = guitarList.iterator();
        	while (itrGuitars.hasNext())
        	{
        		// Reset to standard tuning
        		((Guitar)itrGuitars.next()).resetToStandard(true);
        	}
        }
        else if (inAction.equals("Change Orientation"))
        {
          logger.info("Changing fretboard orientation");
          for(Guitar guitar : guitarList) {
            panel.remove(guitar);
          }
          buildGuitarList(alternateOrientation());
          setContentPane(panel);
          pack();
          setSize(new Dimension(798, 623));
        }
        // Show web browser help
        else if (inAction.equals("Show Help"))
        {
        	logger.info("Launching application help");
        	
        	try
        	{
        		// Launch Help
        		boolean helpAvailable = helpLauncher.showHelp();
        		if (!helpAvailable)
        		{
        			buttonMap.showInfo("No Help");
        		}
        	}
        	catch(Exception ex)
        	{
        		logger.log(Level.SEVERE, "Error Showing Help", ex);
        	}
        }
        // Set Filtering Mode to No Filter
        else if (inAction.equals("No Filter"))
        {
        	logger.info("Selected No Filter");
        	ApplicationModes.setFilterMode(ScaleFilter.NO_FILTER);
        	
        	guitarMenu.setFilterMenu();
        }
        // Set Filtering Mode to 3 Note Per String Filter
        else if (inAction.equals("3 Note Filter"))
        {
        	logger.info("Selected 3 Note Filter");
        	ApplicationModes.setFilterMode(ScaleFilter.THREE_NOTE_FILTER);
        	
        	guitarMenu.setFilterMenu();
        }
        // Set Filtering Mode to Box Pattern Filter
        else if (inAction.equals("Box Filter"))
        {
        	logger.info("Selected Box Filter");
        	ApplicationModes.setFilterMode(ScaleFilter.BOX_FILTER);
        	
        	guitarMenu.setFilterMenu();
        }
        // Play the current scale midi
        else if (inAction.equals("Play Scale"))
        {
        	logger.info("Playing scale");
        	Iterator itrGuitars = guitarList.iterator();
        	while (itrGuitars.hasNext())
        	{
        		Guitar tmpGuitar = (Guitar)itrGuitars.next();
        		Fretboard tmpFretboard = tmpGuitar.getFretboard();
        		if (tmpFretboard.getFretboardType() == Fretboard.DISPLAY)
        		{
        			midiPlayer.setTempo(buttonMap.getTempo());
        			midiPlayer.setVolume(buttonMap.getVolume());
        			midiPlayer.setGuitar(buttonMap.getGuitarType());
        			midiPlayer.playScale(tmpGuitar.getTuningShifts(), tmpFretboard.getScale());
        		}
        	}
        }
        // Play the current scale and chord midi
        else if (inAction.equals("Play Chord"))
        {
        	logger.info("Playing Scale and Chords");
        	Scale clickScale = null;
        	Scale dispScale = null;
        	ArrayList clickShifts = null;
        	ArrayList dispShifts = null;
        	
        	Iterator itrGuitars = guitarList.iterator();
        	while (itrGuitars.hasNext())
        	{
        		Guitar tmpGuitar = (Guitar)itrGuitars.next();
        		Fretboard tmpFretboard = tmpGuitar.getFretboard();
        		if (tmpFretboard.getFretboardType() == Fretboard.DISPLAY)
        		{
        			dispScale = tmpFretboard.getScale();
        			dispShifts = tmpGuitar.getTuningShifts();
        		}
        		else if (tmpFretboard.getFretboardType() == Fretboard.CLICKABLE)
        		{
        			clickScale = tmpFretboard.getScale();
        			clickShifts = tmpGuitar.getTuningShifts();
        		}
        	}
        	
        	midiPlayer.setTempo(buttonMap.getTempo());
        	midiPlayer.setVolume(buttonMap.getVolume());
        	midiPlayer.setGuitar(buttonMap.getGuitarType());
        	midiPlayer.playScaleAndChord(clickShifts, clickScale, dispShifts, dispScale);	
    
        }
        // Search Scales by name
        else if(inAction.equals("Search By Name"))
        {
        	logger.info("Searching for scale name = " + buttonMap.getSearchName());
        	ArrayList scaleList = scaleManager.searchScalesByName(buttonMap.getSearchName());
        	buttonMap.setScaleList(scaleList);
        }
        // Search Scales by name
        else if(inAction.equals("Scale Selected"))
        {
        	logger.info("Selected scale from combo box to view");
        	Scale viewScale = scaleManager.getScale(buttonMap.getSelectedScaleName());
        	if (viewScale != null)
        	{
        		Iterator itrGuitars = guitarList.iterator();
        		while (itrGuitars.hasNext())
        		{
        			((Guitar)itrGuitars.next()).viewScale(viewScale);
        		}
        	}
        }
        // Set the filter position
        else if (inAction.equals("Set Position"))
        {
        	logger.info("Selected Root Position");
        	Iterator itrGuitars = guitarList.iterator();
        	while (itrGuitars.hasNext())
        	{
        		Guitar tmpGuitar = (Guitar)itrGuitars.next();
        		Fretboard tmpFretboard = tmpGuitar.getFretboard();
        		if (tmpFretboard.getFretboardType() == Fretboard.DISPLAY)
        		{
        			tmpGuitar.setFilterPosition(buttonMap.getSelectedPosition());
        		}
        	}
        	
        }
        // Find scales by notes
        // Starts With Find, ends with number of allowable notes
        else if (inAction.startsWith("Find Scales"))
        {
        	logger.info("Searching for scales containing notes");
        	
        	// Set allowable off notes
        	int allowableOffNotes = 0;
        	if (inAction.endsWith("1"))
        	{
        		allowableOffNotes = 1;
        	}
        	else if (inAction.endsWith("2"))
        	{
        		allowableOffNotes = 2;
        	}
        	
        	
        	Iterator itrGuitars = guitarList.iterator();
        	while (itrGuitars.hasNext())
        	{
        		Guitar tmpGuitar = (Guitar)itrGuitars.next();
        		Fretboard tmpFretboard = tmpGuitar.getFretboard();
        		if (tmpFretboard.getFretboardType() == Fretboard.CLICKABLE)
        		{
        			if (tmpFretboard.hasScaleLocations())
        			{
        				// Reset Guitar to standard tuning first
        				tmpGuitar.resetToStandard(false);
        				
        				Scale tmpScale = tmpFretboard.getScale();
        			
        				ArrayList scaleList = scaleManager.searchByLocations(tmpScale, allowableOffNotes);
        				buttonMap.setScaleList(scaleList);
        				
        				logger.info("Found " + String.valueOf(scaleList.size()-1) 
        							+ " scales containing notes allow off notes=" + allowableOffNotes);
        			}
        			else
        			{
        				logger.log(Level.WARNING, "Searching for scale with no notes");
        				
        				// Show No Notes selected info
        				buttonMap.showInfo("No Notes");
        			}
        		}
        	}
        }
        // Set the chord building mode
        else if (inAction.equals("Pick Note Mode"))
        {
        	ApplicationModes.togglePickMode();
        	
        	if (ApplicationModes.pickMode)
        	{
        		logger.info("Chord selection mode");
        		
        		setCursor(Cursor.HAND_CURSOR);
        		buttonMap.showInfo("Pick Mode");
        	}
        	else
        	{
        		logger.info("Return to normal mode");
        		
        		setCursor(Cursor.DEFAULT_CURSOR);
        	}
        }
        // Switch the guitars tuning
        else if (inAction.startsWith("Tune -"))
        {
        	String tuningName = inAction.replaceAll("Tune -", "");
        	
        	ArrayList shiftUnits = (ArrayList)tuningsManager.getTuningShiftUnits(tuningName);
        	if (shiftUnits != null)
        	{
        		logger.info("Tuning guitars");
        		
        		Iterator itrGuitars = guitarList.iterator();
        		while (itrGuitars.hasNext())
        		{
        			// Reset to standard tuning first, then tune using offsets
        			Guitar tmpGuitar = (Guitar)itrGuitars.next();
        			tmpGuitar.resetToStandard(true);
        			tmpGuitar.tuneGuitar(shiftUnits);
        		}
        	}
        }
        // View A Scale
        else
        {
        	Scale viewScale = scaleManager.getScale(inAction);
        	if (viewScale != null)
        	{
        		logger.info("Viewing scale");
        		
        		Iterator itrGuitars = guitarList.iterator();
        		while (itrGuitars.hasNext())
        		{
        			((Guitar)itrGuitars.next()).viewScale(viewScale);
        		}
        	}
        }
        
        repaint();
    }
    
    private Orientation alternateOrientation() {
      Orientation orientation = Orientation.LEFT_HAND;
      if(guitarList.get(0).getFretboardOrientation() == orientation) {
        orientation = Orientation.RIGHT_HAND;
      }
      return orientation;
    }

    // Button and Menu Listener
    public void actionPerformed(ActionEvent e) 
    {
        String action = e.getActionCommand();
        
        if (action.equals("Exit"))
        {
        	logger.info("Exited application");
        	
        	GuitarKeyFrame.this.windowClosed();
        }
        else
        {
        	handleAction(action);
        }
    }
    
    private void switchApplicationMode(int inMode)
    {
    	appModes.setApplicationMode(inMode);
    }
    
    
    /**
     * Shutdown procedure when run as an application.
     */
    protected void windowClosed() 
    {
    	
    	System.exit(0);
    }
    
}

/**
* This class controls the setup for the main panel
*/
class GuitarPanel extends JPanel
{
	Image backgroundImg;
	    
	public GuitarPanel()
	{
	    backgroundImg = new ImageIcon(getClass().getClassLoader().getResource("img/outline.png")).getImage();   
	}
	
    public void paintComponent(Graphics g)
	{
	    if (backgroundImg != null)
        {
            g.drawImage(backgroundImg,0,0,this);
        }
	}
}
