package GuitarKey.Gui;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.AffineTransform;

import javax.swing.ImageIcon;

import GuitarKey.Utils.ApplicationModes;

/**
 * Class implements a turnable knob interface
 */
public class TurnButton extends Component implements MouseListener, MouseMotionListener
{
	// Obj Values
	private int lowValue = 0;
	private int highValue = 100;
	private int currentValue = 50;

	private Image backImg = null;
	private Image turnImg = null;

	// Rotation parameters
	private double rotation = 0.0;
	private double rotationDelta = 3.0;

	private String title = "";

	// x, y, height, width constants
	public static final int TB_HEIGHT = 46; // Component height
	public static final int TB_WIDTH = 78; // Component width
	public static final int TURN_X_OFFSET = 5;
	public static final int TURN_Y_OFFSET = 5;
	public static final int DISPLAY_X_OFFSET = 52;
	public static final int DISPLAY_Y_OFFSET = 31;
	public static final int TITLE_X_OFFSET = 39;
	public static final int TITLE_Y_OFFSET = 10;

	// Mouse x tracking
	private int lastX = 0;
	private int x = 0;
	private int diffX = 0;

	private static final double ROTATION_MAX = 150;

	/**
	 * Constructor
	 */
	public TurnButton(String inName, int minVal, int maxVal)
	{
		super();

		// Images
		ClassLoader cl = getClass().getClassLoader();
		backImg = new ImageIcon(cl.getResource("img/menu/turn_button_back.png")).getImage();
		turnImg = new ImageIcon(cl.getResource("img/menu/turn_button.png")).getImage();

		title = inName;
		lowValue = minVal;
		highValue = maxVal;
		currentValue = (maxVal + minVal)/2;
		rotationDelta = (ROTATION_MAX*2) / (maxVal-minVal);

		addMouseListener(this);
		addMouseMotionListener(this);
	}

	/**
	 * Set the component x, y location
	 */
	public void setLocation(int inX, int inY)
	{
		super.setBounds(inX, inY, TB_WIDTH, TB_HEIGHT);
	}

	/**
	 * Get the current components value
	 */
	public int getValue()
	{
		return currentValue;
	}

	// from the MouseListener interface
    public void mousePressed(MouseEvent e)
    {
        lastX = e.getX();
    }


    // from the MouseListener interface
    public void mouseReleased(MouseEvent e)
    {
    	// reset last x positions
    	lastX = 0;
    }


    // from the MouseListener interface
    public void mouseClicked(MouseEvent e)
    {
        // do nothing
    }


    // from the MouseListener interface
    public void mouseEntered(MouseEvent e)
    {
        // do nothing
    }


    // from the MouseListener interface
    public void mouseExited(MouseEvent e)
    {
        // do nothing
    }

	// from the MouseMotionListener interface
    public void mouseDragged(MouseEvent e)
    {
        // Process mouse drags (only concerned w/ x movement)
        x = e.getX();

        if (lastX != 0)
        {
        	diffX = x - lastX;

        	// Mouse dragged right (increase button)
        	if (diffX > 0)
        	{
        		rotation += rotationDelta; // degrees per turn
        		currentValue++;

        		if (rotation > ROTATION_MAX)
        		{
        			rotation = ROTATION_MAX;
        			currentValue = highValue;
        		}
        	}
        	// Mouse dragged left (decrease button)
        	else if (diffX < 0)
        	{
        		rotation -= rotationDelta; // degrees per turn
        		currentValue--;

        		if (rotation < -ROTATION_MAX)
        		{
        			rotation = -ROTATION_MAX;
        			currentValue = lowValue;
        		}
        	}
        }

        lastX = x;

        repaint();
    }


    // from the MouseMotionListener interface
    public synchronized void mouseMoved(MouseEvent e)
    {
        // do nothing
    }

	// Override paint
	public void paint(Graphics g)
	{
		// Background Image
		g.drawImage(backImg, 0, 0, null);

		// Rotation Image
		Graphics2D g2d = (Graphics2D)g;
		AffineTransform origXform = g2d.getTransform();
		AffineTransform newXform = (AffineTransform)(origXform.clone());
		//center of rotation is center of the button rotation image
		int xRot = TURN_X_OFFSET + 16;
		int yRot = TURN_Y_OFFSET + 16;
		newXform.rotate(Math.toRadians(rotation), xRot, yRot);
		g2d.setTransform(newXform);
		// Draw rotation image
		int x = TURN_X_OFFSET;
		int y = TURN_Y_OFFSET;
		g2d.drawImage(turnImg, x, y, this);
		g2d.setTransform(origXform);

		// Draw Text
		g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    	g2d.setFont(ApplicationModes.FONT_DIGI14);
    	g2d.setColor(ApplicationModes.COLOR_YELLOW);
    	g2d.drawString(String.valueOf(currentValue), DISPLAY_X_OFFSET, DISPLAY_Y_OFFSET);

    	g2d.setFont(ApplicationModes.FONT_GEORGIA10);
    	g2d.setColor(ApplicationModes.COLOR_GRAY);
    	g2d.drawString(title, TITLE_X_OFFSET, TITLE_Y_OFFSET);
	}
}
