package GuitarKey.Instruments;

import java.awt.Image;
import java.awt.Point;
import java.util.Iterator;
import java.util.logging.Logger;

import javax.swing.ImageIcon;

import GuitarKey.Utils.Scale;
import GuitarKey.Utils.ScaleLocation;
import GuitarKey.Utils.ShiftUnit;

/**
 * Class holds a fretboard object
 */
public class Fretboard extends FretboardLocationMap
{
	// Logger
	private static Logger logger = Logger.getLogger("GuitarKey");

	// Type of fretboard, clickable or display-only
	public static final int CLICKABLE = 0;
	public static final int DISPLAY = 1;

	private int fretboardType = 0;

	// Fretboard images
	private Image fretboardImg = null;
	private Image clickableFretboardImg = null;
	private Image displayFretboardImg = null;
	private Image clickNoteImg = null;
	private Image clickNoteSelImg = null;
	private Image displayNoteImg = null;
	private Image displayRootNoteImg = null;
	private Image scaleNameImg = null;

	// The Scale associated with this fretboard
	Scale scale;


	/**
	 * Constructor
	 */
	public Fretboard(int fretboardType, Orientation orientation)
	{
		super(fretboardType, orientation);

		logger.info("Creating Fretboard");

		this.fretboardType = fretboardType;

		ClassLoader cl = getClass().getClassLoader();

		if(orientation == Orientation.LEFT_HAND) {
  		displayFretboardImg = new ImageIcon(cl.getResource("img/neck2.png")).getImage();
      clickableFretboardImg = new ImageIcon(cl.getResource("img/neck_click2.png")).getImage();
		} else {
		  displayFretboardImg = new ImageIcon(cl.getResource("img/neck2_rh.png")).getImage();
		  clickableFretboardImg = new ImageIcon(cl.getResource("img/neck_click2_rh.png")).getImage();
		}

    clickNoteImg = new ImageIcon(cl.getResource("img/note_c.png")).getImage();
    clickNoteSelImg = new ImageIcon(cl.getResource("img/note_s.png")).getImage();
		displayNoteImg = new ImageIcon(cl.getResource("img/note_d.png")).getImage();
		displayRootNoteImg = new ImageIcon(cl.getResource("img/note_r.png")).getImage();

		scaleNameImg = new ImageIcon(cl.getResource("img/scale_name.png")).getImage();

        if (fretboardType == CLICKABLE)
		{
			logger.info("Fretboard Type = Clickable");
			fretboardImg = clickableFretboardImg;
		}
		else if (fretboardType == DISPLAY)
		{
			logger.info("Fretboard Type = Display");
			fretboardImg = displayFretboardImg;
		}

		// Temporary blank scale
		scale = new Scale("C", "Unknown");
	}


	/**
	 * Set the fretboard type and change the image accordingly
	 */
	public void setFretboardType(int fretboardType)
	{
		this.fretboardType = fretboardType;

		if (fretboardType == CLICKABLE)
		{
			fretboardImg = clickableFretboardImg;
		}
		else if (fretboardType == DISPLAY)
		{
			fretboardImg = displayFretboardImg;
		}
	}

	/**
	 * Method gets the fretboard type
	 */
	public int getFretboardType()
	{
		return fretboardType;
	}

	/**
	 * Method gets the fretboard width
	 */
	public int getWidth()
	{
		return WIDTH;
	}

	/**
	 * Method sets the fretboard scale to the input scale
	 */
	public void setScale(Scale inScale)
	{
		logger.info("Setting fretboard with " + inScale.getScaleFullName() + " scale");

		scale = inScale;
	}

	/**
	 * Method returns the current scale
	 */
	public Scale getScale()
	{
		return scale;
	}

	/**
	 * Method returns the scale back to a blank scale
	 */
	public void resetScale()
	{
		logger.info("Reseting fretboard scale");

		// Reset the scale
		scale = new Scale("C", "New Scale");
	}

	/**
	 * Method adds a scale location to this fretboards scale
	 */
	public void addScaleLocation(ScaleLocation inScaleLoc)
	{
		scale.addScaleLocation(inScaleLoc);
	}

	/**
	 * Method removes a scale location from this fretboards scale
	 */
	public void removeScaleLocation(ScaleLocation inScaleLoc)
	{
		scale.removeScaleLocation(inScaleLoc);
	}

	/**
	 * Method tunes the fretboard based on the ShiftUnit (string, shift)
	 */
	public void tune(ShiftUnit inShiftUnit)
	{
		// Cycle through scale locations
		Iterator itrScaleLocs = scale.getScaleLocations().iterator();
		while (itrScaleLocs.hasNext())
		{
			ScaleLocation tmpScaleLoc = (ScaleLocation)itrScaleLocs.next();

			// If string being tuned matches the scale locations string, tune
			if (tmpScaleLoc.getStringLocation() == inShiftUnit.getStringNum())
			{
				int shift = inShiftUnit.getShift();

				if (shift != 0)
				{
					logger.info("Tuning string " + inShiftUnit.getStringNum() + " by " + shift + " units");
				}

				if (shift > 0)
				{
					tmpScaleLoc.increase(shift);
				}
				else if (shift < 0)
				{
					tmpScaleLoc.decrease(shift);
				}
			}
		}
	}

	/**
	 * Method returns the fretboard image
	 */
	public Image getFretboardImage()
	{
		return fretboardImg;
	}

	/**
	 * Method returns a clicked note image
	 */
	public Image getClickNoteImage()
	{
		return clickNoteImg;
	}

	/**
	 * Method returns a clicked note selected image
	 */
	public Image getClickNoteSelectedImage()
	{
		return clickNoteSelImg;
	}

	/**
	 * Method returns a displayed note image
	 */
	public Image getDisplayNoteImage()
	{
		return displayNoteImg;
	}

	/**
	 * Method returns a displayed note image
	 */
	public Image getDisplayRootNoteImage()
	{
		return displayRootNoteImg;
	}

	/**
	 * Method returns the scale name backdrop image
	 */
	public Image getScaleNameImage()
	{
		return scaleNameImg;
	}

	/**
	 * Method returns whether fretboard has any scale locations
	 */
	public boolean hasScaleLocations()
	{
		return !(scale.getScaleLocations().isEmpty());
	}

	/**
	 * Method tests whether fretboard contains the clicked point
	 */
	public boolean containsPoint(Point inPoint)
	{
		boolean containsPoint = false;

		int pY = (int)inPoint.getY();
		int pX = transformX((int)inPoint.getX());
		
    	if ((pY > getY()) && (pY < (getY()+HEIGHT))
    	 && (pX > getX()) && (pX < (getX()+WIDTH)))
    	{
    		containsPoint = true;
    	}

    	return containsPoint;
	}
}
