package GuitarKey.Instruments;

import java.awt.Point;
import java.util.logging.Logger;

import GuitarKey.Utils.ApplicationModes;
import GuitarKey.Utils.Note;
import GuitarKey.Utils.ScaleLocation;

/**
 * Class defines the locations of allowable notes on the fretboard
 */
public class FretboardLocationMap
{
	// Logger
	private static Logger logger = Logger.getLogger("GuitarKey");

	public static enum Orientation {LEFT_HAND, RIGHT_HAND};
	  
  private Orientation orientation = Orientation.LEFT_HAND;
	  
	// Slopes for strings to determine x,y coordinates
	public static final float E6_STRING_SLOPE = 0.0227618f;
	public static final float A5_STRING_SLOPE = 0.0122061f;
	public static final float D4_STRING_SLOPE = 0.0015129f;
	public static final float G3_STRING_SLOPE = -0.0075643f;
	public static final float B2_STRING_SLOPE = -0.0162242f;
	public static final float E1_STRING_SLOPE = -0.0287443f;

	// Offset values used to determine x,y coordinate (b value)
	public static final float E6_STRING_OFFSET = 12.606217f;
	public static final float A5_STRING_OFFSET = 36.803109f;
	public static final float D4_STRING_OFFSET = 59.950777f;
	public static final float G3_STRING_OFFSET = 82.098574f;
	public static final float B2_STRING_OFFSET = 104.270725f;
	public static final float E1_STRING_OFFSET = 127.443006f;

	private int[] frets;
	private float[] slopes;
	private float[] offsets;
	private int numOfFrets;

	private int x;
	private int y;

	// X, Y values
	private double xVal;
	private double yVal;
	
  // Fretboard pixel height and width
  public static final int HEIGHT = 140;
  public static final int WIDTH = 700;

	/**
	 * Constructor
	 */
	public FretboardLocationMap(int fretboardType, Orientation orientation)
	{
	  this.orientation = orientation;
	  
		logger.info("Creating FreboardLocationMap");

		// Setup array with fret x locations
		numOfFrets = ApplicationModes.NUMBER_OF_FRETS; // including zero so 24 = 22
		frets = new int[]{700, 688, 635, 587, 539, 495, 454, 415, 378, 344, 311, 279, 250, 221, 195, 171, 148, 125, 105, 85, 67, 49, 33, 17};

		slopes = new float[]{E6_STRING_SLOPE, A5_STRING_SLOPE, D4_STRING_SLOPE,
								G3_STRING_SLOPE, B2_STRING_SLOPE, E1_STRING_SLOPE};

		offsets = new float[]{E6_STRING_OFFSET, A5_STRING_OFFSET, D4_STRING_OFFSET,
								G3_STRING_OFFSET, B2_STRING_OFFSET, E1_STRING_OFFSET};

	}

	// Set the fretboard x location
	public void setX(int inX)
	{
		x = inX;
	}

	// Set the fretboard y location
	public void setY(int inY)
	{
		y = inY;
	}

	// Get the fretboard x location
	public int getX()
	{
		return x;
	}

	// Get the fretboard y location
	public int getY()
	{
		return y;
	}


	/**
	 * Method returns the closest Scale location for a point
	 */
	public ScaleLocation getScaleLocation(Point inPoint)
	{
		xVal = transformX(inPoint.getX());
		yVal = inPoint.getY();

		int fretLoc = 0;
		int stringLoc = 0;

		// Last Fret out of range but still on fretboard
		if (xVal < (frets[numOfFrets-1]+x))
		{
			fretLoc = numOfFrets-2;
		}

		// Determine the X location
		for (int i=1; i<numOfFrets; i++)
		{
			if (((frets[i-1]+x) > xVal) && ((frets[i]+x) <= xVal))
			{
				// Midway between two frets
				fretLoc = i-1;
			}
		}

		// Determine the string location
		stringLoc = getStringNumber(xVal, yVal);

    xVal = transformX(xVal);

		// Create the return point
		ScaleLocation scaleLocation = new ScaleLocation(stringLoc, fretLoc);

		return scaleLocation;
	}


	/**
	 * Method returns the Note (x,y) location of a scale location
	 */
	public Note getNoteLocation(ScaleLocation inScaleLoc)
	{
		int xLoc = 0;
		int yLoc = 0;

		int stringNum = inScaleLoc.getStringLocation();
		int fretNum = inScaleLoc.getFretLocation();

		if (inScaleLoc.isAboveOctave())
		{
			// Set to above octave
			fretNum = fretNum + ApplicationModes.NUMBER_OF_NOTES_IN_OCTAVE;
		}

		if (fretNum <= 22)
		{
			xLoc = (int)(((frets[fretNum]+x) + (frets[fretNum+1]+x))/2);
		}

		if (xLoc > 0)
		{
			yLoc = getYPosFromString(xLoc, stringNum);
		}

		xLoc = transformX(xLoc);

		Note note = new Note(xLoc, yLoc);
		note.setRoot(inScaleLoc.isRoot());
		note.setSelected(inScaleLoc.isSelected());
		
		return note;
	}



	/**
	 * Method calculates the string of the note click
	 */
	private int getStringNumber(double inX, double inY)
	{
		// Cycle through slopes and offsets to calculate the closest string
		int string = 1;
		float fLoc;
		float tempLoc;

		float diff = 0.0f, diff2 = 0.0f;

		// Y = MX + B
		fLoc = slopes[0] * (float)inX + offsets[0] + y;
		diff = Math.abs(fLoc - (float)inY);

		for (int i=1; i<6; i++)
		{
			tempLoc = slopes[i] * (float)inX + offsets[i] + y; // Y = MX + B
			diff2 = Math.abs(tempLoc - (float)inY);

			if (diff2 < diff)
			{
				diff = diff2;
				string = i+1;
			}
		}

		return string;
	}

	/**
	 * Method calculates the Y position of the note
	 */
	private int getYPosFromString(double inX, int stringNum)
	{
		// Use stringNumber to reference slope
		int iLoc = 0;
		float fLoc;


		// Y = MX + B
		fLoc = slopes[stringNum-1] * (float)inX + offsets[stringNum-1] + y;


		iLoc = (int)fLoc;

		return iLoc;
	}

  public Orientation getOrientation() {
    return orientation;
  }

  protected int transformX(int x) {
    int newX = x;
    if(orientation == Orientation.RIGHT_HAND) {
      newX =  WIDTH - x;
    }
    return newX;
  }

  private double transformX(double x) {
    double newX = x;
    if(orientation == Orientation.RIGHT_HAND) {
      newX =  WIDTH - x;
    }
    return newX;
  }
}
