package GuitarKey.Instruments;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Component;
import java.awt.Composite;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.logging.Logger;

import GuitarKey.Filters.ScaleFilter;
import GuitarKey.Instruments.FretboardLocationMap.Orientation;
import GuitarKey.Utils.ApplicationModes;
import GuitarKey.Utils.Note;
import GuitarKey.Utils.Scale;
import GuitarKey.Utils.ScaleLocation;
import GuitarKey.Utils.ShiftUnit;

/**
 * Class implements a guitar interface.  Controls click processing and holds fretboard and tuning map data.
 */
public class Guitar extends Component implements MouseListener
{
	// Logger
	private static Logger logger = Logger.getLogger("GuitarKey");

	// Fretboard and Tuning Map objects
	private Fretboard fretboard;
	private TuningMap tuningMap;

	public static final int GUITAR_WIDTH = 760;
	public static final int GUITAR_HEIGHT = 155;

	private int GUITAR_Y_OFFSET = 15;

	// Padding between tuner map beginnin of guitar neck
	private int TUNER_MAPX_PADDING = 705;

	private Image imageNote;

	// Scale Filter
	private ScaleFilter scaleFilter;

	///////// REMOVE - DEBUG SLOPE CALC /////////////
	//Point lastPoint = new Point(0, 0);
	/////////////////////////////////////////////////


	/**
	 * Constructor
	 */
	public Guitar(int fretboardType, Orientation orientation)
	{
		super();

		logger.info("Creating Guitar");

		fretboard = new Fretboard(fretboardType, orientation);
		tuningMap = new TuningMap();
		scaleFilter = new ScaleFilter();

		setBounds(0, 0, GUITAR_WIDTH, GUITAR_HEIGHT);

		addMouseListener(this);
	}


    /**
     * Method overrides the setLocation method
     */
    public void setLocation(int inX, int inY)
    {
    	super.setLocation(inX, inY);
    	// Set fretboard and tuning map relative to guitar location
    	fretboard.setX(0);
    	tuningMap.setX(TUNER_MAPX_PADDING);

    	// Set fretboard and tuning map relative to guitar location
    	fretboard.setY(GUITAR_Y_OFFSET);
    	tuningMap.setY(GUITAR_Y_OFFSET);
    }


    /**
     * Method clears the notes of the clickable fretboards
     */
    public void clearNotes()
    {
    	if (fretboard.getFretboardType() == Fretboard.CLICKABLE)
    	{
    		fretboard.resetScale();
    	}
    }

    /**
     * Method sets the viewable guitar scale
     */
    public void viewScale(Scale inScale)
    {
    	if (fretboard.getFretboardType() == Fretboard.DISPLAY)
    	{
    		fretboard.setScale(inScale);

    		// Adjust for tuning if fretboard is alt. tuned
    		ArrayList shiftUnits = tuningMap.getTuningShifts();

	    	Iterator itrShiftUnits = shiftUnits.iterator();
	    	while (itrShiftUnits.hasNext())
	    	{
	    		ShiftUnit shiftUnit = (ShiftUnit)itrShiftUnits.next();
	    		fretboard.tune(shiftUnit);
	    	}
    	}

    	repaint();
    }

    /**
	 * Method returns the guitar to standard tuning
	 */
    public void resetToStandard(boolean bothGuitars)
    {
    	logger.info("Reseting guitar " + fretboard.getFretboardType() + " to standard tuning");

    	// Reset the tuning map and get the array list of shift units
    	if (fretboard.getFretboardType() == Fretboard.CLICKABLE)
    	{
	    	ArrayList shiftUnits = tuningMap.resetToStandard();

	    	Iterator itrShiftUnits = shiftUnits.iterator();
	    	while (itrShiftUnits.hasNext())
	    	{
	    		ShiftUnit shiftUnit = (ShiftUnit)itrShiftUnits.next();
	    		fretboard.tune(shiftUnit);
	    	}
    	}
    	// Reset display fretboard if flag is set
    	if (bothGuitars)
    	{
    		// Reset the tuning map and get the array list of shift units
    		if (fretboard.getFretboardType() == Fretboard.DISPLAY)
    		{
	    		ArrayList shiftUnits = tuningMap.resetToStandard();

	    		Iterator itrShiftUnits = shiftUnits.iterator();
	    		while (itrShiftUnits.hasNext())
	    		{
	    			ShiftUnit shiftUnit = (ShiftUnit)itrShiftUnits.next();
	    			fretboard.tune(shiftUnit);
	    		}
    		}
    	}
    }

    /**
     * Method gets the tuning shifts for this guitar
     */
    public ArrayList getTuningShifts()
    {
    	return tuningMap.getTuningShifts();
    }

    /**
     * Method tunes the clickable Fretboard by the amount of tuning shifts
     */
    public void tuneClickableGuitar(ArrayList inTuningShifts)
    {
    	if (fretboard.getFretboardType() == Fretboard.CLICKABLE)
    	{
    		Iterator itrShiftUnits = inTuningShifts.iterator();
			while (itrShiftUnits.hasNext())
			{
				ShiftUnit shiftUnit = (ShiftUnit)itrShiftUnits.next();
				fretboard.tune(shiftUnit);
				tuningMap.tune(shiftUnit);
			}
		}
    }

    /**
     * Method tunes the Fretboard by the amount of tuning shifts
     */
    public void tuneGuitar(ArrayList inTuningShifts)
    {
    	Iterator itrShiftUnits = inTuningShifts.iterator();
		while (itrShiftUnits.hasNext())
		{
			ShiftUnit shiftUnit = (ShiftUnit)itrShiftUnits.next();
			fretboard.tune(shiftUnit);
			tuningMap.tune(shiftUnit);
		}
    }

    /**
     * Method sets the pattern filter position for the scales
     */
    public void setFilterPosition(String inPosition)
    {
    	scaleFilter.setFilterPosition(inPosition);
    }

    /**
     * Method gets the guitars fretboard
     */
    public Fretboard getFretboard()
    {
    	return fretboard;
    }


    // from the KeyListener interface
    public void keyPressed(KeyEvent e)
    {
        // make sure the key isn't processed for anything else
        e.consume();
    }


    // from the KeyListener interface
    public void keyReleased(KeyEvent e)
    {
        // make sure the key isn't processed for anything else
        e.consume();
    }


    // from the KeyListener interface
    public void keyTyped(KeyEvent e)
    {
        // make sure the key isn't processed for anything else
        e.consume();
    }


    // from the MouseListener interface
    public void mousePressed(MouseEvent e)
    {

    }


    /**
     * Method processes the mouse release on the fretboard and tuner map
     */
    public void mouseReleased(MouseEvent e)
    {
    	// Process mouse released point
    	if (e.getButton() == MouseEvent.BUTTON1)
    	{
    		Point clickPoint = e.getPoint();

    		// Test fretboard click
    		if (fretboard.containsPoint(clickPoint) && fretboard.getFretboardType() == Fretboard.CLICKABLE)
    		{
    			// Get the scale location and add it to the fretboard
    			ScaleLocation scaleLoc = fretboard.getScaleLocation(clickPoint);
    			fretboard.addScaleLocation(scaleLoc);

    			////////// REMOVE - DEBUG SLOPE CALC ////////////
    			//Float slope;
    			//slope = ((float)clickPoint.getY() - (float)lastPoint.getY()) / ((float)clickPoint.getX() - (float)lastPoint.getX());
    			//System.out.println("Slope for string " + scaleLoc.getStringLocation() + " = " + slope);
    			//lastPoint = clickPoint;
    			/////////////////////////////////////////////////
    		}
    		else if (tuningMap.containsPoint(clickPoint))
    		{
    			ShiftUnit shiftUnit = tuningMap.processClick(clickPoint);

    			// Tune fretboard
    			fretboard.tune(shiftUnit);
    		}
    	}
    	// Process Right Click Release
    	else if (e.getButton() == MouseEvent.BUTTON3)
    	{
    		Point clickPoint = e.getPoint();

    		// Test fretboard click
    		if (fretboard.containsPoint(clickPoint) && fretboard.getFretboardType() == Fretboard.CLICKABLE)
    		{
    			// Get the scale location and remove it from the fretboard
    			ScaleLocation scaleLoc = fretboard.getScaleLocation(clickPoint);
    			fretboard.removeScaleLocation(scaleLoc);
    		}
    	}

    	repaint();
    }


    // from the MouseListener interface
    public void mouseClicked(MouseEvent e)
    {
        // do nothing
    }


    // from the MouseListener interface
    public void mouseEntered(MouseEvent e)
    {
        // do nothing
    }


    // from the MouseListener interface
    public void mouseExited(MouseEvent e)
    {
        // do nothing
    }

    public void drawTranslucent(Graphics2D g2d, Image img, int x, int y, float translucency)
	{
		int width = Note.NOTE_OFFSET * 2;
		Composite c = g2d.getComposite();
		g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,translucency));
		g2d.drawImage(img, x, y, null);
		g2d.setColor(new Color(255, 255, 255, 200));
		g2d.fillRoundRect(x, y, width, width, width, width);
		g2d.setComposite(c);
	}


    /**
     * Method draws the guitar objects
     */
    public void paint(Graphics g)
    {
    	// Font properties
    	Graphics2D gfx = (Graphics2D)g;
    	gfx.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    	g.setFont(ApplicationModes.FONT_UBER);
    	g.setColor(ApplicationModes.COLOR_RED);

    	// Draw the fretboard
    	g.drawImage(fretboard.getFretboardImage(), fretboard.getX(), fretboard.getY(), null);

    	// Process guitars based on type
    	// DISPLAY
    	if (fretboard.getFretboardType() == Fretboard.DISPLAY)
    	{
    		// Scale Name if fretboard has notes
    		if (fretboard.hasScaleLocations())
    		{
    			g.drawImage(fretboard.getScaleNameImage(),
    					fretboard.getX() + fretboard.getWidth()/2, 14, null);

    			g.drawString(fretboard.getScale().getScaleFullName().replace('#', 's').toLowerCase(),
    					fretboard.getX() + fretboard.getWidth()/2 + 10, 16);
    		}

    		// Notes from fretboard to draw
    		// Display Fretboard
    		Scale scale = fretboard.getScale();

    		// Apply Filters
    		if (fretboard.hasScaleLocations())
    		{
    			scaleFilter.filter(scale);
    		}

    		Iterator itrNotes = scale.getScaleLocations().iterator();
	    	while (itrNotes.hasNext())
	    	{
	    		ScaleLocation tmpScaleLoc = (ScaleLocation)itrNotes.next();
	    		// Draw regular note
	    		Note tmpNote = fretboard.getNoteLocation(tmpScaleLoc);
	    		if (tmpNote.isRoot())
		    	{
		    		if (!tmpScaleLoc.alphaApplied())
		    		{
		    			g.drawImage(fretboard.getDisplayRootNoteImage(), tmpNote.getX(), tmpNote.getY(), null);
		    		}
		    		else
		    		{
		    			drawTranslucent((Graphics2D)g, fretboard.getDisplayRootNoteImage(),
		    								tmpNote.getX(), tmpNote.getY(), 0.3F);
		    		}
		    	}
		    	else
		    	{
		    		if (!tmpScaleLoc.alphaApplied())
		    		{
		    			g.drawImage(fretboard.getDisplayNoteImage(), tmpNote.getX(), tmpNote.getY(), null);
		    		}
		    		else
		    		{
		    			drawTranslucent((Graphics2D)g, fretboard.getDisplayNoteImage(),
		    								tmpNote.getX(), tmpNote.getY(), 0.3F);
		    		}
		    	}
		    	// Draw note one octave higher
		    	tmpNote = fretboard.getNoteLocation(tmpScaleLoc.getAboveOctave());
		    	if (tmpNote.isAvailable())
		    	{
		    		if (tmpNote.isRoot())
			    	{
			    		if (!tmpScaleLoc.alphaApplied())
		    			{
			    			g.drawImage(fretboard.getDisplayRootNoteImage(), tmpNote.getX(), tmpNote.getY(), null);
			    		}
			    		else
		    			{
		    				drawTranslucent((Graphics2D)g, fretboard.getDisplayRootNoteImage(),
		    									tmpNote.getX(), tmpNote.getY(), 0.3F);
		    			}
			    	}
			    	else
			    	{
			    		if (!tmpScaleLoc.alphaApplied())
		    			{
		    				g.drawImage(fretboard.getDisplayNoteImage(), tmpNote.getX(), tmpNote.getY(), null);
		    			}
		    			else
			    		{
			    			drawTranslucent((Graphics2D)g, fretboard.getDisplayNoteImage(),
			    								tmpNote.getX(), tmpNote.getY(), 0.3F);
			    		}
			    	}
		    	}
    		}
    	}
    	// CLICKABLE
    	else
    	{
    		// Notes from fretboard to draw
    		// Clickable Fretboard
	    	Iterator itrNotes = fretboard.getScale().getScaleLocations().iterator();
	    	while (itrNotes.hasNext())
	    	{
	    		Note tmpNote = fretboard.getNoteLocation((ScaleLocation)itrNotes.next());
				if (tmpNote.isSelected())
				{
					g.drawImage(fretboard.getClickNoteSelectedImage(), tmpNote.getX(), tmpNote.getY(), null);
				}
				else
				{
					g.drawImage(fretboard.getClickNoteImage(), tmpNote.getX(), tmpNote.getY(), null);
				}
	    	}
    	}

    	// Tuning Map and Tuners
    	Iterator itrTuners = tuningMap.getTuners().iterator();
    	while (itrTuners.hasNext())
    	{
    		Tuner tmpTuner = (Tuner)itrTuners.next();
    		g.drawImage(tmpTuner.getNoteImage(), tmpTuner.getNoteX(), tmpTuner.getY(), null);
    		g.drawImage(tmpTuner.getLeftClickImage(), tmpTuner.getLeftClickX(), tmpTuner.getY(), null);
    		g.drawImage(tmpTuner.getRightClickImage(), tmpTuner.getRightClickX(), tmpTuner.getY(), null);
    	}
    }
    
    public Orientation getFretboardOrientation() {
      return fretboard.getOrientation();
    }
}
