package GuitarKey.Instruments;

import java.awt.Image;
import java.awt.Point;

import javax.swing.ImageIcon;

import GuitarKey.Utils.ApplicationModes;
import GuitarKey.Utils.ShiftUnit;

/**
 * Class holds a tuner object
 */
class Tuner
{
	// Tuner string number and note Index
	private int stringNumber;
	private int tunerIndex;

	private int x;
	private int y;

	// Drawing X Offsets
	private int NOTE_X_OFFSET = 12;
	private int LEFT_X_OFFSET = 0;
	private int RIGHT_X_OFFSET = 41;

	private int WIDTH = 51;
	private int HEIGHT = 23;

	Image[] noteImgs;

	private Image noteImage;
	private Image leftClickImg;
	private Image rightClickImg;

	private int NUMBER_OF_NOTES;


	/**
	 * Constructor
	 * input argument is Guitar String number
	 */
	public Tuner(int inStringNum)
	{
		stringNumber = inStringNum;

		NUMBER_OF_NOTES = ApplicationModes.NUMBER_OF_AVAILABLE_NOTES;

		noteImgs = new Image[NUMBER_OF_NOTES];

		loadImages();

		leftClickImg = new ImageIcon(getClass().getClassLoader().getResource("img/tuner/left.png")).getImage();
		rightClickImg = new ImageIcon(getClass().getClassLoader().getResource("img/tuner/right.png")).getImage();

		tunerIndex = ApplicationModes.Standard_Tuning_Index[stringNumber-1];
	}

	/**
	 * Method loads the tuning note images array
	 */
	private void loadImages()
	{
		ClassLoader cl = getClass().getClassLoader();

		for (int i=0; i<NUMBER_OF_NOTES; i++)
		{
			noteImgs[i] = new ImageIcon(cl.getResource("img/tuner/" + ApplicationModes.Note_Lower_Case_Names[i] + ".png")).getImage();
		}
	}

	/**
	 * Method to set the current tuner string number
	 */
	public void setStringNumber(int inStringNum)
	{
		stringNumber = inStringNum;

		// Resets back to standard tuning for this string
		tunerIndex = ApplicationModes.Standard_Tuning_Index[stringNumber-1];
	}

	/**
	 * Method returns this tuners string number
	 */
	public int getStringNumber()
	{
		return stringNumber;
	}

	// Set the tuner x location
	public void setX(int inX)
	{
		x = inX;
	}

	// Set the tuner y location
	public void setY(int inY)
	{
		y = inY;
	}

	// Get the tuner x location
	public int getX()
	{
		return x;
	}

	// Get the tuner y location
	public int getY()
	{
		return y;
	}

	// Get the tuner height
	public int getHeight()
	{
		return HEIGHT;
	}

	// Get the tuner width
	public int getWidth()
	{
		return WIDTH;
	}

	/**
	 * Get the current note image
	 */
	public Image getNoteImage()
	{
		return noteImgs[tunerIndex];
	}

	/**
	 * Get the current note image x offset
	 */
	public int getNoteX()
	{
		return (x+NOTE_X_OFFSET);
	}

	/**
	 * Get the current left click image
	 */
	public Image getLeftClickImage()
	{
		return leftClickImg;
	}

	/**
	 * Get the current left click image x offset
	 */
	public int getLeftClickX()
	{
		return (x+LEFT_X_OFFSET);
	}

	/**
	 * Get the current right click image
	 */
	public Image getRightClickImage()
	{
		return rightClickImg;
	}

	/**
	 * Get the current right click image x offset
	 */
	public int getRightClickX()
	{
		return (x+RIGHT_X_OFFSET);
	}

	/**
	 * Get the current note name
	 */
	public String getNoteName()
	{
		return ApplicationModes.Note_Lower_Case_Names[tunerIndex];
	}

	/**
	 * Method processes the click point on this tuner
	 * return the shift (number of frets) of the tuning
	 * and the string number in a ShiftUnit
	 */
	public ShiftUnit processClick(Point inPoint)
	{
		// shift offset (number of frets)
		int shift = 0;

		// Since point is already inside tuner, only process x coordinate
		int pX = (int)inPoint.getX();

		// Clicked left button
		if (pX < (x+NOTE_X_OFFSET))
		{
			shift = tuneDown();
		}
		// Clicked right button
		else if (pX > (x+RIGHT_X_OFFSET))
		{
			shift = tuneUp();
		}

		return (new ShiftUnit(stringNumber, shift));
	}

	/**
	 * Method tunes this tuner index up
	 * returns strings shift in tuning
	 */
	public int tuneUp()
	{
		int shift = -1; // used for note offset (opposite direction of tune)

		tunerIndex++;

		if (tunerIndex >= NUMBER_OF_NOTES)
		{
			tunerIndex = 0; // Reset to zero if cycled passed end
		}

		return shift;
	}

	/**
	 * Method tunes this tuner index down
	 * returns strings shift in tuning
	 */
	public int tuneDown()
	{
		int shift = 1; // used for note offset (opposite direction of tune)

		tunerIndex--;

		if (tunerIndex < 0)
		{
			tunerIndex = NUMBER_OF_NOTES - 1; // Reset to end if cycled before beginning
		}

		return shift;
	}

	/**
	 * Method returns the cumulative shift for this tuning peg
	 * used for returning clickable fretboard back to standard tuning
	 */
	public ShiftUnit getStandardTuneShift()
	{
		int cumulativeShift = tunerIndex - ApplicationModes.Standard_Tuning_Index[stringNumber-1];
		return (new ShiftUnit(stringNumber, cumulativeShift));
	}

	/**
	 * Method returns the cumulative shift for this tuning peg
	 * used for setting display fretboard back to alt tuning
	 */
	public ShiftUnit getCumulativeShift()
	{
		int cumulativeShift = ApplicationModes.Standard_Tuning_Index[stringNumber-1] - tunerIndex;
		return (new ShiftUnit(stringNumber, cumulativeShift));
	}

	/**
	 * Method returns the tuner to standard tuning for this string
	 */
	public void tuneStandard()
	{
		tunerIndex = ApplicationModes.Standard_Tuning_Index[stringNumber-1]; // reset tuner index
	}

	/**
	 * Method tests whether tuner contains the clicked point
	 */
	public boolean containsPoint(Point inPoint)
	{
		boolean containsPoint = false;

		int pY = (int)inPoint.getY();
		int pX = (int)inPoint.getX();

    	if ((pY > y) && (pY < (y+HEIGHT))
    	 && (pX > x) && (pX < (x+WIDTH)))
    	{
    		containsPoint = true;
    	}

    	return containsPoint;
	}
}
