package GuitarKey.Sound;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MetaEventListener;
import javax.sound.midi.MetaMessage;
import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Sequence;
import javax.sound.midi.Sequencer;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Synthesizer;
import javax.sound.midi.Track;

import GuitarKey.Utils.ApplicationModes;
import GuitarKey.Utils.Scale;
import GuitarKey.Utils.ScaleLocation;
import GuitarKey.Utils.ShiftUnit;


/**
 * Class handles all note midi sounds
 */
public class MidiPlayer
{
	// Logger
	private static Logger logger = Logger.getLogger("GuitarKey");

	// 52 - 121
	public static final int END_OF_TRACK = 47;

	public static final int ACOUSTIC_GUITAR = 24;
	public static final int JAZZ_GUITAR = 26;
	public static final int ELECTRIC_GUITAR = 30;

	public static final int MID_TEMPO = 170;
	public static final int MID_VOLUME = 50;
	public static final int MID_FREQUENCY = 4;

	public static final int BASE_NOTE = 28;

	// Editable properties
	private int instrument;
	private int tempo;
	private int velocity;
	private int chordVelocity;
	private int chordFrequency;

	private boolean playing = false;

	// Fixed objects
	private int[] noteSequence;
	private int[] chordNoteSequence;
	private int[] scaleNoteSequence;
	private Sequence sequence;
	private Sequencer sequencer;
	private Synthesizer synthesizer;
	private Track track;
	private ShortMessage sm;
	private ShortMessage on;
	private ShortMessage off;

    /**
     * Constructor
     */
	public MidiPlayer()
	{
		logger.info("Creating MidiPlayer");

		instrument = ACOUSTIC_GUITAR;
		tempo = MID_TEMPO;
		velocity = MID_VOLUME;
		chordVelocity = (velocity * 3)/4;
		chordFrequency = MID_FREQUENCY;

		try
		{
			// Set up the Sequencer and Synthesizer objects
			sequencer = MidiSystem.getSequencer( );
			sequencer.open( );
			synthesizer = MidiSystem.getSynthesizer( );
			synthesizer.open( );
			sequencer.getTransmitter( ).setReceiver(synthesizer.getReceiver( ));
		}
		catch (Exception ex)
		{
			logger.log(Level.SEVERE, "Error creating sequencer and synthesizer", ex);
		}
	}

	/**
	 * Set the player tempo
	 */
	public void setTempo(int inTempo)
	{
		tempo = inTempo;
	}

	/**
	 * Set the player volume
	 */
	public void setVolume(int inVolume)
	{
		velocity = inVolume;

		chordVelocity = (velocity * 3)/4;
	}

	/**
	 * Set the instrument guitar type
	 */
	public void setGuitar(int inType)
	{
		instrument = inType;
	}

	/**
	 * Play the scale based on the note list passed in
	 */
	public void playScale(ArrayList shiftUnits, Scale inScale)
	{
		if (!playing)
		{
			playing = true;
			noteSequence = buildNoteArray(shiftUnits, inScale);

			try
			{
				play(noteSequence);
			}
			catch (Exception ex){ ex.printStackTrace(); }
		}
	}

	/**
	 * Play the scale and chords based on notes passed in
	 */
	public void playScaleAndChord(ArrayList chordShift, Scale chordScale,
										ArrayList scaleShift, Scale inScale)
	{
		if (!playing)
		{
			playing = true;
			chordNoteSequence = buildChordNoteArray(chordShift, chordScale);
			scaleNoteSequence = buildNoteArray(scaleShift, inScale);

			try
			{
				play(chordNoteSequence, scaleNoteSequence);
			}
			catch (Exception ex){ ex.printStackTrace(); }
		}
	}

	// Scale
	private void play(int[] notes)
		throws MidiUnavailableException, InvalidMidiDataException
	{
        // 16 ticks per quarter note.
        sequence = new Sequence(Sequence.PPQ, 16);

        // Add the specified notes to the track
        addTrack(sequence, instrument, tempo, notes);

		// Specify the sequence to play, and the tempo to play it at
		sequencer.setSequence(sequence);
		sequencer.setTempoInBPM(tempo);

		// Let us know when it is done playing
		sequencer.addMetaEventListener(new MetaEventListener( ) {
			public void meta(MetaMessage m)
			{
				// A message of this type is automatically sent
				// when we reach the end of the track
				if (m.getType( ) == END_OF_TRACK)
				{
					playing = false;
				}
			}
		});

		// And start playing now.
		sequencer.start( );
	}

	// Scale and Chord
	private void play(int[] chordNotes, int[] scaleNotes)
		throws MidiUnavailableException, InvalidMidiDataException
	{
        // 16 ticks per quarter note.
        sequence = new Sequence(Sequence.PPQ, 16);

        // Add the specified notes to the track
        addTrack(sequence, instrument, tempo, chordNotes, scaleNotes);

		// Specify the sequence to play, and the tempo to play it at
		sequencer.setSequence(sequence);
		sequencer.setTempoInBPM(tempo);

		// Let us know when it is done playing
		sequencer.addMetaEventListener(new MetaEventListener( ) {
			public void meta(MetaMessage m)
			{
				// A message of this type is automatically sent
				// when we reach the end of the track
				if (m.getType( ) == END_OF_TRACK)
				{
					playing = false;
				}
			}
		});

		// And start playing now.
		sequencer.start( );
	}

	/*
     * This method parses the specified int[  ] of notes into a Track.
     */
    private void addTrack(Sequence s, int instrument, int tempo,
                                int[] notes)
        throws InvalidMidiDataException
    {
        track = s.createTrack( );  // Begin with a new track

        // Set the instrument on channel 0
        sm = new ShortMessage( );
        sm.setMessage(ShortMessage.PROGRAM_CHANGE, 0, instrument, 0);
        track.add(new MidiEvent(sm, 0));

        int n = 0; // current character in notes[  ] array
        int t = 0; // time in ticks for the composition

        // These values persist and apply to all notes 'till changed
        int notelength = 16; // default to quarter notes
        boolean sustain = false;   // is the sustain pedal depressed?
        int numnotes = 0;    // How many notes in current chord?

        while(n < notes.length)
        {
            int key = notes[n];
            n++;

			addNote(track, t, notelength, key, velocity);
			numnotes++;

			// Now add additional rest time
			t += notelength;
        }
    }

    /*
     * This method parses the specified int[] of notes and int[] of chord into a Track.
     */
    private void addTrack(Sequence s, int instrument, int tempo,
                                int[] chord, int [] notes)
        throws InvalidMidiDataException
    {
        track = s.createTrack( );  // Begin with a new track

        // Set the instrument on channel 0
        sm = new ShortMessage( );
        sm.setMessage(ShortMessage.PROGRAM_CHANGE, 0, instrument, 0);
        track.add(new MidiEvent(sm, 0));

        int n = 0; // current character in notes[  ] array
        int t = 0; // time in ticks for the composition
        int c = 0; // counter frequency counter

        // These values persist and apply to all notes 'till changed
        int notelength = 16; // default to quarter notes
        boolean sustain = false;   // is the sustain pedal depressed?
        int numnotes = 0;    // How many notes in current chord?

        while(n < notes.length)
        {
            int key = notes[n];
            n++;

            numnotes = 0;

			addNote(track, t, notelength, key, velocity);

			if (c == 0)
			{
				while (numnotes < chord.length)
				{
					addNote(track, t, notelength*chordFrequency, chord[numnotes], chordVelocity);
					numnotes++;
				}
			}

			c++;
			if (c >= chordFrequency)
			{
				c=0;
			}

			// Now add additional rest time
			t += notelength;
        }

        // No scale available, play chord
        if (notes.length == 0)
        {
        	numnotes = 0;

        	while (numnotes < chord.length)
			{
				addNote(track, t, notelength*chordFrequency, chord[numnotes], velocity);
				numnotes++;
			}
        }
    }

    // A convenience method to add a note to the track on channel 0
    private void addNote(Track track, int startTick,
                               int tickLength, int key, int velocity)
        throws InvalidMidiDataException
    {
        on = new ShortMessage( );
        on.setMessage(ShortMessage.NOTE_ON,  0, key, velocity);
        off = new ShortMessage( );
        off.setMessage(ShortMessage.NOTE_OFF, 0, key, velocity);
        track.add(new MidiEvent(on, startTick));
        track.add(new MidiEvent(off, startTick + tickLength));
    }

    /**
     * Method gets the string array of note names from the scale based on the string shifts
     */
    private int[] buildNoteArray(ArrayList shiftUnits, Scale inScale)
    {
    	ArrayList noteList = new ArrayList();
    	int lastNote = 0;

    	int scaleInts[];

    	inScale.sortScaleLocations();

    	Iterator itrScaleLocs = inScale.getScaleLocations().iterator();
    	while (itrScaleLocs.hasNext())
    	{
    		ScaleLocation tmpScaleLoc = (ScaleLocation)itrScaleLocs.next();

    		Iterator itrShifts = shiftUnits.iterator();
    		while (itrShifts.hasNext())
    		{
    			ShiftUnit tmpShift = (ShiftUnit)itrShifts.next();

    			// String Numbers match
    			if (tmpScaleLoc.getStringLocation() == tmpShift.getStringNum())
    			{
    				// Standard tuning index
    				int index = ApplicationModes.Standard_Tuning_Index[tmpScaleLoc.getStringLocation() - 1];

    				// - tuning offset
    				index = index - tmpShift.getShift();

    				// + fret location
    				index = index + tmpScaleLoc.getFretLocation();

    				// Note int
    				if (index > lastNote)
    				{
    					noteList.add(String.valueOf(index + BASE_NOTE));
    					lastNote = index;
    				}
    			}
    		}
    	}

    	scaleInts = new int[noteList.size()];
    	for (int i=0; i<scaleInts.length; i++)
    	{
    		scaleInts[i] = Integer.parseInt((String)noteList.get(i));
    	}

    	return scaleInts;
    }

    /**
     * Method gets the string array of note names from the chord based on the string shifts
     */
    private int[] buildChordNoteArray(ArrayList shiftUnits, Scale inScale)
    {
    	ArrayList noteList = new ArrayList();
    	int lastNote = 0;

    	int scaleInts[];

    	inScale.sortScaleLocations();

    	Iterator itrScaleLocs = inScale.getScaleLocations().iterator();
    	while (itrScaleLocs.hasNext())
    	{
    		ScaleLocation tmpScaleLoc = (ScaleLocation)itrScaleLocs.next();

    		Iterator itrShifts = shiftUnits.iterator();
    		while (itrShifts.hasNext())
    		{
    			ShiftUnit tmpShift = (ShiftUnit)itrShifts.next();

    			// String Numbers match
    			if (tmpScaleLoc.getStringLocation() == tmpShift.getStringNum())
    			{
    				// Standard tuning index
    				int index = ApplicationModes.Standard_Tuning_Index[tmpScaleLoc.getStringLocation() - 1];

    				// - tuning offset
    				index = index - tmpShift.getShift();

    				// + fret location
    				index = index + tmpScaleLoc.getFretLocation();

    				// Add selected chord notes
    				if (tmpScaleLoc.isSelected())
    				{
    					noteList.add(String.valueOf(index + BASE_NOTE));
    					lastNote = index;
    				}
    			}
    		}
    	}

    	scaleInts = new int[noteList.size()];
    	for (int i=0; i<scaleInts.length; i++)
    	{
    		scaleInts[i] = Integer.parseInt((String)noteList.get(i));
    	}

    	return scaleInts;
    }
}
