package GuitarKey.Utils;

import java.io.File;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Class manages all the scales reading and writing of scales
 */
public class ScaleManager
{
	// Logger
	private static Logger logger = Logger.getLogger("GuitarKey");

	private HashMap scaleMap;
	private ArrayList searchList;
	private ArrayList offNotes;


	/**
	 * Scale Manager constructor
	 */
	public ScaleManager()
	{
		logger.info("Creating ScaleManager");

		scaleMap = new HashMap(512);
		searchList = new ArrayList();
		offNotes = new ArrayList();
		buildScaleMap();
	}

	// Build the scales map
	private void buildScaleMap()
	{
		logger.info("Building Scale Map");

		File xmlFiles = new File("data/keydata");

		String[] fileList = xmlFiles.list(new XMLFilter());

		int length = 0;
		int fretNum = 0;

		Document document;

		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();

		// Cycle through File List
		for (int i=0; i<fileList.length; i++)
		{
			try
			{
				// Parse the XML file
   				DocumentBuilder builder = factory.newDocumentBuilder();
   				document = builder.parse( new File("data/keydata/" + fileList[i]) );

   				Element root = document.getDocumentElement();

   				String rootNote = root.getElementsByTagName("RootNote").item(0).getFirstChild().getNodeValue();
   				String scaleName = root.getElementsByTagName("Name").item(0).getFirstChild().getNodeValue();

   				Scale tmpScale = new Scale(rootNote, scaleName);

   				// Get the string notes
   				for (int j=1; j<=6; j++)
   				{
   					NodeList stringList = root.getElementsByTagName("String" + j);
   					// Make sure string has notes
   					if (stringList.getLength() > 0)
   					{
   						NodeList fretList = ((Element)stringList.item(0)).getElementsByTagName("Fret");
   						length = fretList.getLength();

   						for (int k=0; k<length; k++)
   						{
   							fretNum = Integer.parseInt(fretList.item(k).getFirstChild().getNodeValue());
   							ScaleLocation scaleLoc = new ScaleLocation(j, fretNum);
   							boolean isRoot = isRootNote(rootNote, scaleLoc);
   							scaleLoc.setRoot(isRoot);
   							tmpScale.addScaleLocation(scaleLoc);
   						}
   					}
   				}

   				// set the scale's note names
   				setNoteNames(tmpScale);

   				// Add the scale to the scale map by the full name
   				scaleMap.put(tmpScale.getScaleFullName(), tmpScale);

			}
			catch (Exception spe)
			{
				logger.log(Level.SEVERE, "Error building scale map from xml files", spe);
			}
		}
	}

	/**
	 * Add a scale to the scale manager and write the file
	 * returns if manager already has scale
	 */
	public boolean addScale(Scale inScale, boolean saveToFile)
	{
		logger.info("Attempting to save scale");

		boolean containsScale = scaleMap.containsKey(inScale.getScaleFullName());

		// Set the note names
		setNoteNames(inScale);

		scaleMap.put(inScale.getScaleFullName(), inScale);

		if (saveToFile)
		{
			try
			{
				// Save this scale to a file
				File outFile = new File("data/keydata/" + escapeCharacters(inScale.getScaleFullName()) + ".xml");
				FileWriter out = new FileWriter(outFile);

				Document document = null;
				DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
				DocumentBuilder builder = factory.newDocumentBuilder();
				DOMImplementation impl = builder.getDOMImplementation();

   				Element e;
   				Element e2;
   				Node n;

   				document = impl.createDocument(null, "GuitarKey", null);
				Element root = document.getDocumentElement();

   				// Root Note
   				e = document.createElementNS(null, "RootNote");
  				n = document.createTextNode(inScale.getScaleRoot());
  				e.appendChild(n);
  				root.appendChild(e);

  				// Scale Name
   				e = document.createElementNS(null, "Name");
  				n = document.createTextNode(inScale.getScaleName());
  				e.appendChild(n);
  				root.appendChild(e);

  				// Notes
  				String lastStr = null;
  				String strNum;
  				String fretNum;

  				Element note_e = document.createElementNS(null, "Notes");

  				Iterator itrScaleLocs = inScale.getScaleLocations().iterator();
  				while (itrScaleLocs.hasNext())
  				{
  					ScaleLocation tmpScaleLoc = (ScaleLocation)itrScaleLocs.next();
  					strNum = String.valueOf(tmpScaleLoc.getStringLocation());
  					fretNum = String.valueOf(tmpScaleLoc.getFretLocation());
  					if (!strNum.equals(lastStr))
  					{
  						if (lastStr != null)
  						{
  							// Close last string
  							note_e.appendChild(e);
  						}
  						// New String
  						e = document.createElementNS(null, "String" + strNum);
  					}
  					e2 = document.createElementNS(null, "Fret");
  					n = document.createTextNode(fretNum);

  					// Add notes
  					e2.appendChild(n);
  					e.appendChild(e2);


  					lastStr = strNum;
  				}
  				note_e.appendChild(e);
  				root.appendChild(note_e);


  				// Write To File
				DOMSource domSource = new DOMSource(document);
				StreamResult streamResult = new StreamResult(out);
				TransformerFactory tf = TransformerFactory.newInstance();
				Transformer serializer = tf.newTransformer();
				serializer.setOutputProperty(OutputKeys.ENCODING,"UTF-8");
				serializer.setOutputProperty(OutputKeys.INDENT,"yes");
				serializer.transform(domSource, streamResult);

   			}
   			catch (Exception spe)
   			{
   				logger.log(Level.SEVERE, "Error saving scale xml file", spe);
   			}
		}

		return containsScale;
	}

	/**
	 * Get the list of available scales
	 */
	public List getScaleNames()
	{
		List scaleNameList = new ArrayList(scaleMap.keySet());

		Collections.sort(scaleNameList);

		return scaleNameList;
	}

	/**
	 * Get the scale by scale name
	 */
	public Scale getScale(String scaleName)
	{
		Scale returnScale = (Scale)scaleMap.get(scaleName);
		if (returnScale != null)
		{
			returnScale = returnScale.clone();
		}

		return returnScale;
	}


	/**
	 * Search the available scales by name
	 */
	public ArrayList searchScalesByName(String inName)
	{
		// Build the list of scale names matching search
		String scaleName;
		searchList.clear();

		Iterator itrScaleNames = getScaleNames().iterator();
		while (itrScaleNames.hasNext())
		{
			scaleName = (String)itrScaleNames.next();
			if (scaleName.toLowerCase().contains(inName.toLowerCase()))
			{
				searchList.add(scaleName);
			}
		}

		return searchList;
	}

	/**
	 * Search the available scales by matching scale locations
	 */
	public ArrayList searchByLocations(Scale inScale, int allowableOffNotes)
	{
		// Build the list of scale names matching search
		String scaleName;

		searchList.clear();

		setNoteNames(inScale);

		Iterator itrScaleNames = getScaleNames().iterator();
		while (itrScaleNames.hasNext())
		{
			boolean addScale = true;
			offNotes.clear();

			scaleName = (String)itrScaleNames.next();
			Scale scale = (Scale)scaleMap.get(scaleName);

			Iterator itrNoteNames = inScale.getNoteNameList().iterator();
			while (itrNoteNames.hasNext())
			{
				String noteName = (String)itrNoteNames.next();

				if (!scale.contains(noteName))
				{
					// Found location not in scale, add to allowable off notes
					if (!offNotes.contains(noteName))
					{
						offNotes.add(noteName);
					}
				}
			}

			// Check to see if off notes are more than allowed
			if (offNotes.size() > allowableOffNotes)
			{
				addScale = false;
			}

			// add scale to list
			if (addScale)
			{
				searchList.add(scaleName);
			}
		}

		return searchList;
	}

	/**
	 * Method tests whether a scale location is a root for a scale
	 */
	private boolean isRootNote(String inRoot, ScaleLocation inScaleLoc)
	{
		boolean result = false;

		int stringNum = inScaleLoc.getStringLocation();
		int fretNum = inScaleLoc.getFretLocation();

		// Standard Tuning index seed for this string
		int idx = ApplicationModes.Standard_Tuning_Index[stringNum-1];

		String note0 = ApplicationModes.Note_Upper_Case_Names[idx];
		int note0Idx = 0;
		int rootNoteIdx = 0;

		// Cycle through note names and determine fret numbers of root
		for (int i=0; i<ApplicationModes.NUMBER_OF_NOTES_IN_OCTAVE; i++)
		{
			if (inRoot.equalsIgnoreCase(ApplicationModes.Note_Names[i]))
			{
				rootNoteIdx = i;
			}
			if(note0.equalsIgnoreCase(ApplicationModes.Note_Names[i]))
			{
				note0Idx = i;
			}
		}

		int difference = rootNoteIdx - note0Idx;

		if (difference < 0)
		{
			// See if fret is root or octave above
			int rootFret = ApplicationModes.NUMBER_OF_NOTES_IN_OCTAVE + difference;
			if (fretNum == rootFret || fretNum == (rootFret+ApplicationModes.NUMBER_OF_NOTES_IN_OCTAVE))
			{
				result = true;
			}
		}
		else if (difference >= 0)
		{
			// See if fret is root or octave above
			int rootFret = difference;
			if (fretNum == rootFret || fretNum == (rootFret+ApplicationModes.NUMBER_OF_NOTES_IN_OCTAVE))
			{
				result = true;
			}
		}

		return result;
	}

	/**
	 * Method sets the note names for the scale
	 */
	public void setNoteNames(Scale inScale)
	{
		int stringNum;
		int fretNum;
		int idx;
		String noteName;

		Iterator itrScaleLocations = inScale.getScaleLocations().iterator();
		while (itrScaleLocations.hasNext())
		{
			ScaleLocation tmpScaleLoc = (ScaleLocation)itrScaleLocations.next();

			stringNum = tmpScaleLoc.getStringLocation();
			fretNum = tmpScaleLoc.getFretLocation();

			// Standard Tuning index seed for this string
			idx = ApplicationModes.Standard_Tuning_Index[stringNum-1];

			idx += fretNum;
			if (idx >= ApplicationModes.NUMBER_OF_AVAILABLE_NOTES)
			{
				idx -= ApplicationModes.NUMBER_OF_AVAILABLE_NOTES;
			}

			noteName = ApplicationModes.Note_Upper_Case_Names[idx];

			tmpScaleLoc.setNoteName(noteName);
		}
	}

	/**
	 * Method escapes the characters of an input string to remove characters who have misbehaved
	 * in this case the # sign because new File() w/ a # sign sucks
	 */
	private String escapeCharacters(String inString)
	{
		return (inString.replace('#', 's'));
	}

//	//////////////////////////////////////////////////////////
//	public void fixScales()
//	{
//		Iterator itrScaleNames = getScaleNames().iterator();
//		while (itrScaleNames.hasNext())
//		{
//			boolean addScale = true;
//
//			String scaleName = (String)itrScaleNames.next();
//			Scale scale = (Scale)scaleMap.get(scaleName);
//
//			addScale(scale, true);
//		}
//	}
//	//////////////////////////////////////////////////////////

	// XML Filter
	class XMLFilter implements FilenameFilter
	{
		public boolean accept(File dir, String name)
		{
        	return (name.endsWith(".xml"));
        }
    }
}
