package GuitarKey.Utils;

import java.io.File;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Class manages all the tunings, reading and writing of tunings files
 */
public class TuningsManager
{
	// Logger
	private static Logger logger = Logger.getLogger("GuitarKey");

	private HashMap tuningsMap;


	/**
	 * Tunings Manager constructor
	 */
	public TuningsManager()
	{
		logger.info("Creating TuningsManager");

		tuningsMap = new HashMap(256);
		buildTuningsMap();
	}

	// Build the tunings map
	private void buildTuningsMap()
	{
		logger.info("Building Tunings Map");

		File xmlFiles = new File("data/shiftdata");

		String[] fileList = xmlFiles.list(new XMLFilter());

		Document document;

		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();

		// Cycle through File List
		for (int i=0; i<fileList.length; i++)
		{
			try
			{
				// Parse the XML file
   				DocumentBuilder builder = factory.newDocumentBuilder();
   				document = builder.parse( new File("data/shiftdata/" + fileList[i]) );

   				Element root = document.getDocumentElement();

   				String tuningName = root.getElementsByTagName("Name").item(0).getFirstChild().getNodeValue();
   				int stringNumber;
   				int tuningShift;

   				ArrayList shiftUnits = new ArrayList();

				NodeList shiftsList = root.getElementsByTagName("Shifts");

				for (int j=0; j<shiftsList.getLength(); j++)
				{
					NodeList shiftList = ((Element)shiftsList.item(j)).getElementsByTagName("Shift");

					for (int k=0; k<shiftList.getLength(); k++)
					{
						Element shiftElement = (Element)shiftList.item(k);

						// Get the String Number
						NodeList stringList = shiftElement.getElementsByTagName("String");
						stringNumber = Integer.parseInt(stringList.item(0).getFirstChild().getNodeValue());

						// Get the tuning shift
						NodeList offsetList = shiftElement.getElementsByTagName("Offset");
						tuningShift = Integer.parseInt(offsetList.item(0).getFirstChild().getNodeValue());

						// Add the Shift Unit to the Shift Unit array list
						shiftUnits.add(new ShiftUnit(stringNumber, tuningShift));
					}
				}


   				// Add the shift unit list to the tunings map by its name
   				tuningsMap.put(tuningName, shiftUnits);

			}
			catch (Exception spe)
			{
				logger.log(Level.SEVERE, "Error building shifts map from xml files", spe);
			}
		}
	}

	/**
	 * Add a tuning to the tunings manager and write the file
	 * returns if manager already has tuning
	 */
	public boolean addTuning(String inTuningName, ArrayList inShiftUnits, boolean saveToFile)
	{
		logger.info("Attempting to save tuning");

		boolean containsTuning = tuningsMap.containsKey(inTuningName);

		tuningsMap.put(inTuningName, inShiftUnits);

		if (saveToFile)
		{
			try
			{
				// Save this tuning to a file
				File outFile = new File("data/shiftdata/" + escapeCharacters(inTuningName) + ".xml");
				FileWriter out = new FileWriter(outFile);

				Document document = null;
				DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
				DocumentBuilder builder = factory.newDocumentBuilder();
				DOMImplementation impl = builder.getDOMImplementation();

   				Element e;
   				Element e2;
   				Element e3;
   				Node n;
   				Node n2;

   				document = impl.createDocument(null, "Tuning", null);
				Element root = document.getDocumentElement();

   				// Tuning Name
   				e = document.createElementNS(null, "Name");
  				n = document.createTextNode(inTuningName);
  				e.appendChild(n);
  				root.appendChild(e);

  				// Notes
  				String strNum;
  				String offset;

  				Element shifts_e = document.createElementNS(null, "Shifts");

  				Iterator itrShiftUnits = inShiftUnits.iterator();
  				while (itrShiftUnits.hasNext())
  				{
  					ShiftUnit tmpShiftUnit = (ShiftUnit)itrShiftUnits.next();
  					strNum = String.valueOf(tmpShiftUnit.getStringNum());
  					offset = String.valueOf(tmpShiftUnit.getShift());

  					e = document.createElementNS(null, "Shift");

  					e2 = document.createElementNS(null, "String");
  					n = document.createTextNode(strNum);

  					e3 = document.createElementNS(null, "Offset");
  					n2 = document.createTextNode(offset);

  					// Add Shifts
  					e2.appendChild(n);
  					e3.appendChild(n2);
  					e.appendChild(e2);
  					e.appendChild(e3);

  					shifts_e.appendChild(e);
  				}
  				root.appendChild(shifts_e);


  				// Write To File
				DOMSource domSource = new DOMSource(document);
				StreamResult streamResult = new StreamResult(out);
				TransformerFactory tf = TransformerFactory.newInstance();
				Transformer serializer = tf.newTransformer();
				serializer.setOutputProperty(OutputKeys.ENCODING,"UTF-8");
				serializer.setOutputProperty(OutputKeys.INDENT,"yes");
				serializer.transform(domSource, streamResult);

   			}
   			catch (Exception spe)
   			{
   				logger.log(Level.SEVERE, "Error saving tuning xml file", spe);
   			}
		}

		return containsTuning;
	}

	/**
	 * Get the list of available tunings
	 */
	public List getTuningNames()
	{
		List tuningNameList = new ArrayList(tuningsMap.keySet());

		Collections.sort(tuningNameList);

		return tuningNameList;
	}

	/**
	 * Get the tuning shift unit array list by tuning name
	 */
	public ArrayList getTuningShiftUnits(String inTuningName)
	{
		ArrayList result = (ArrayList)tuningsMap.get(inTuningName);

		return result;
	}

	/**
	 * Method escapes the characters of an input string to remove characters who have misbehaved
	 * in this case the # sign because new File() w/ a # sign sucks
	 */
	private String escapeCharacters(String inString)
	{
		return (inString.replace('#', 's'));
	}

	// XML Filter
	class XMLFilter implements FilenameFilter
	{
		public boolean accept(File dir, String name)
		{
        	return (name.endsWith(".xml"));
        }
    }
}
